@@ -31,12 +31,12 @@
   "localhost")
 
 (defun get-host-by-address (address)
-  (with-mapped-conditions ()
+  (with-mapped-conditions (nil address)
     (let ((hostent (posix:resolve-host-ipaddr (host-to-hostname address))))
       (posix:hostent-name hostent))))
 
 (defun get-hosts-by-name (name)
-  (with-mapped-conditions ()
+  (with-mapped-conditions (nil name)
     (let ((hostent (posix:resolve-host-ipaddr name)))
       (mapcar #'host-to-vector-quad
               (posix:hostent-addr-list hostent)))))
@@ -102,7 +102,7 @@
           (let ((*package* (find-package "KEYWORD")))
             (car (read-from-string s t nil :start pos1 :end (1+ pos2)))))))))
 
-(defun handle-condition (condition &optional (socket nil))
+(defun handle-condition (condition &optional (socket nil) (host-or-ip nil))
   "Dispatch a usocket condition instead of a CLISP specific one, if we can."
   (let ((errno
           (cond
@@ -114,13 +114,19 @@
              (car (simple-condition-format-arguments condition))))))
     (when errno
       (let ((error-keyword (if (keywordp errno) errno #+ffi(os:errno errno))))
-        (let ((usocket-error (cdr (assoc error-keyword +clisp-error-map+))))
-          (when usocket-error
-            (if (subtypep usocket-error 'error)
-              (error  usocket-error :socket socket)
-              (signal usocket-error :socket socket))))))))
+        (let ((usock-error (cdr (assoc error-keyword +clisp-error-map+))))
+          (when usock-error
+            (if (subtypep usock-error 'error)
+                (cond ((subtypep usock-error 'ns-error)
+                       (error usock-error :socket socket :host-or-ip host-or-ip))
+                      (t
+                       (error usock-error :socket socket)))
+                (cond ((subtypep usock-error 'ns-condition)
+                       (signal usock-error :socket socket :host-or-ip host-or-ip))
+                      (t
+                       (signal usock-error :socket socket))))))))))
 
-(defun socket-connect (host port &key (protocol :stream) (element-type 'character)
+(defun socket-connect-internal (host &key port (protocol :stream) (element-type 'character)
                        timeout deadline (nodelay t nodelay-specified)
                        local-host local-port)
   (declare (ignorable timeout local-host local-port))
@@ -132,7 +138,7 @@
     (:stream
      (let ((socket)
 	   (hostname (host-to-hostname host)))
-       (with-mapped-conditions (socket)
+       (with-mapped-conditions (socket host)
 	 (setf socket
 	       (if timeout
 		   (socket:socket-connect port hostname
@@ -146,15 +152,16 @@
 			   :stream socket))) ;; the socket is a stream too
     (:datagram
      #+(or rawsock ffi)
-     (socket-create-datagram (or local-port *auto-port*)
-			     :local-host (or local-host *wildcard-host*)
-			     :remote-host (and host (host-to-vector-quad host))
-			     :remote-port port)
+     (with-mapped-conditions (nil (or host local-host))
+       (socket-create-datagram (or local-port *auto-port*)
+                               :local-host (or local-host *wildcard-host*)
+                               :remote-host (and host (host-to-vector-quad host))
+                               :remote-port port))
      #-(or rawsock ffi)
      (unsupported '(protocol :datagram) 'socket-connect))))
 
-(defun socket-listen (host port
-                           &key reuseaddress
+(defun socket-listen-internal
+                          (host &key port reuseaddress
                            (reuse-address nil reuse-address-supplied-p)
                            (backlog 5)
                            (element-type 'character))
@@ -166,7 +173,7 @@
                                    :backlog backlog)
                              (when (ip/= host *wildcard-host*)
                                (list :interface host))))))
-    (with-mapped-conditions ()
+    (with-mapped-conditions (nil host)
         (make-stream-server-socket sock :element-type element-type))))
 
 (defmethod socket-accept ((socket stream-server-usocket) &key element-type)
@@ -184,14 +191,18 @@
 (defmethod socket-close ((usocket usocket))
   "Close socket."
   (with-mapped-conditions (usocket)
-    (close (socket usocket))))
+    (close (socket usocket) :abort t))) ; see "sbcl.lisp" for (:abort t)
 
 (defmethod socket-close ((usocket stream-server-usocket))
   (socket:socket-server-close (socket usocket)))
 
 (defmethod socket-shutdown ((usocket stream-usocket) direction)
   (with-mapped-conditions (usocket)
-    (socket:socket-stream-shutdown (socket usocket) direction)))
+    (if (eq :io direction)
+        (progn
+          (socket:socket-stream-shutdown (socket usocket) :input)
+          (socket:socket-stream-shutdown (socket usocket) :output))
+        (socket:socket-stream-shutdown (socket usocket) direction))))
 
 (defmethod get-local-name ((usocket stream-usocket))
   (multiple-value-bind
