@@ -19,16 +19,22 @@
   (append +unix-errno-condition-map+
           +unix-errno-error-map+))
 
-(defun cmucl-map-socket-error (err &key condition socket)
-  (let ((usock-err
+(defun cmucl-map-socket-error (err &key condition socket host-or-ip)
+  (let ((usock-error
          (cdr (assoc err +cmucl-error-map+ :test #'member))))
-    (if usock-err
-        (if (subtypep usock-err 'error)
-            (error usock-err :socket socket)
-          (signal usock-err :socket socket))
-      (error 'unknown-error
-             :socket socket
-             :real-error condition))))
+    (if usock-error
+        (if (subtypep usock-error 'error)
+            (cond ((subtypep usock-error 'ns-error)
+                   (error usock-error :socket socket :host-or-ip host-or-ip))
+                  (t
+                   (error usock-error :socket socket)))
+            (cond ((subtypep usock-error 'ns-condition)
+                   (signal usock-error :socket socket :host-or-ip host-or-ip))
+                  (t
+                   (signal usock-error :socket socket))))
+        (error 'unknown-error
+               :socket socket
+               :real-error condition))))
 
 ;; CMUCL error handling is brain-dead: it doesn't preserve any
 ;; information other than the OS error string from which the
@@ -40,14 +46,15 @@
 ;;
 ;;
 ;; Just catch the errors and encapsulate them in an unknown-error
-(defun handle-condition (condition &optional (socket nil))
+(defun handle-condition (condition &optional (socket nil) (host-or-ip nil))
   "Dispatch correct usocket condition."
   (typecase condition
     (ext::socket-error (cmucl-map-socket-error (ext::socket-errno condition)
                                                :socket socket
-                                               :condition condition))))
+                                               :condition condition
+                                               :host-or-ip host-or-ip))))
 
-(defun socket-connect (host port &key (protocol :stream) (element-type 'character)
+(defun socket-connect-internal (host &key port (protocol :stream) (element-type 'character)
                        timeout deadline (nodelay t nodelay-specified)
 		       (local-host nil local-host-p)
 		       (local-port nil local-port-p)
@@ -72,7 +79,7 @@
 		 (nconc args (list :local-host (when local-host
 						 (host-to-hbo local-host))
 				   :local-port local-port)))
-	       (with-mapped-conditions (socket)
+	       (with-mapped-conditions (socket host)
 		 (apply #'ext:connect-to-inet-socket args))))
        (if socket
 	   (let* ((stream (sys:make-fd-stream socket :input t :output t
@@ -92,16 +99,16 @@
 		     (nconc args (list :local-host (when local-host
 						     (host-to-hbo local-host))
 				       :local-port local-port)))
-		   (with-mapped-conditions (socket)
+		   (with-mapped-conditions (socket (or host local-host))
 		     (apply #'ext:connect-to-inet-socket args)))
 		 (if (or local-host-p local-port-p)
-		     (with-mapped-conditions (socket)
+		     (with-mapped-conditions (socket (or host local-host))
 		       (apply #'ext:create-inet-listener
 			      (nconc (list (or local-port 0) protocol)
 				     (when (and local-host-p
 						(ip/= local-host *wildcard-host*))
 				       (list :host (host-to-hbo local-host))))))
-		     (with-mapped-conditions (socket)
+		     (with-mapped-conditions (socket (or host local-host))
 		       (ext:create-inet-socket protocol)))))
        (if socket
 	   (let ((usocket (make-datagram-socket socket :connected-p (and host port t))))
@@ -111,14 +118,14 @@
 	   (let ((err (unix:unix-errno)))
 	     (when err (cmucl-map-socket-error err))))))))
 
-(defun socket-listen (host port
-                           &key reuseaddress
+(defun socket-listen-internal
+                          (host &key port reuseaddress
                            (reuse-address nil reuse-address-supplied-p)
                            (backlog 5)
                            (element-type 'character))
  (let* ((reuseaddress (if reuse-address-supplied-p reuse-address reuseaddress))
         (server-sock
-         (with-mapped-conditions ()
+         (with-mapped-conditions (nil host)
            (apply #'ext:create-inet-listener
                   (nconc  (list port :stream
                                 :backlog backlog
@@ -143,7 +150,7 @@
 (defmethod socket-close ((usocket stream-usocket))
   "Close socket."
   (with-mapped-conditions (usocket)
-    (close (socket-stream usocket))))
+    (close (socket-stream usocket) :abort t))) ; see "sbcl.lisp" for (:abort t)
 
 (defmethod socket-close ((usocket usocket))
   "Close socket."
@@ -171,13 +178,14 @@
   (with-mapped-conditions (usocket)
     (ext:inet-shutdown (socket usocket) (ecase direction
                                           (:input ext:shut-rd)
-                                          (:output ext:shut-wr)))))
+                                          (:output ext:shut-wr)
+                                          (:io ext:shut-rdwr)))))
 
 (defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0)
 			&aux (real-buffer (if (zerop offset)
 					      buffer
 					      (subseq buffer offset (+ offset size)))))
-  (with-mapped-conditions (usocket)
+  (with-mapped-conditions (usocket host)
     (if (and host port)
 	(ext:inet-sendto (socket usocket) real-buffer size (host-to-hbo host) port)
 	#-unicode
@@ -237,7 +245,7 @@
                             '((1 ns-host-not-found-error)     ;; HOST_NOT_FOUND
                               (2 ns-no-recovery-error)        ;; NO_DATA
                               (3 ns-no-recovery-error)        ;; NO_RECOVERY
-                              (4 ns-try-again-condition)))))) ;; TRY_AGAIN
+                              (4 ns-try-again-error))))))     ;; TRY_AGAIN
         (when exception
           (error exception))))))
 
@@ -245,13 +253,13 @@
 (defun get-host-by-address (address)
   (handler-case (ext:host-entry-name
                  (lookup-host-entry (host-byte-order address)))
-    (condition (condition) (handle-condition condition))))
+    (condition (condition) (handle-condition condition address))))
 
 (defun get-hosts-by-name (name)
   (handler-case (mapcar #'hbo-to-vector-quad
                         (ext:host-entry-addr-list
                          (lookup-host-entry name)))
-    (condition (condition) (handle-condition condition))))
+    (condition (condition) (handle-condition condition name))))
 
 (defun get-host-name ()
   (unix:unix-gethostname))
