@@ -4,9 +4,7 @@
 
 (defparameter *backend* :iolib)
 
-(eval-when (:load-toplevel :execute)
-  (shadowing-import 'iolib/sockets:socket-option)
-  (export 'socket-option))
+(export 'socket-option)
 
 (defparameter +iolib-error-map+
  `((iolib/sockets:socket-address-in-use-error        . address-in-use-error)
@@ -38,7 +36,7 @@
    ;; Nameservice errors (src/sockets/dns/conditions.lisp)
    (iolib/sockets:resolver-error                     . ns-error)
    (iolib/sockets:resolver-fail-error                . ns-host-not-found-error)
-   (iolib/sockets:resolver-again-error               . ns-try-again-condition)
+   (iolib/sockets:resolver-again-error               . ns-try-again-error)
    (iolib/sockets:resolver-no-name-error             . ns-no-recovery-error)
    (iolib/sockets:resolver-unknown-error             . ns-unknown-error)
    ))
@@ -66,24 +64,21 @@
 	 (usock-error (if (functionp usock-error)
 			  (funcall usock-error condition)
                           usock-error)))
-    (if usock-error
-        (if (typep usock-error 'socket-error)
-            (cond ((subtypep usock-error 'ns-error)
-                  (error usock-error :socket socket :host-or-ip host-or-ip))
-                 (t
-                  (error usock-error :socket socket)))
-            (cond ((subtypep usock-error 'ns-condition)
-                  (signal usock-error :socket socket :host-or-ip host-or-ip))
-                 (t
-                  (signal usock-error :socket socket))))
-        (error 'unknown-error
-               :real-error condition
-               :socket socket))))
+    (when usock-error
+      (if (typep usock-error 'socket-error)
+          (cond ((subtypep usock-error 'ns-error)
+                 (error usock-error :socket socket :host-or-ip host-or-ip))
+                (t
+                 (error usock-error :socket socket)))
+          (cond ((subtypep usock-error 'ns-condition)
+                 (signal usock-error :socket socket :host-or-ip host-or-ip))
+                (t
+                 (signal usock-error :socket socket)))))))
 
 (defun ipv6-address-p (host)
   (iolib/sockets:ipv6-address-p (iolib/sockets:ensure-hostname host)))
 
-(defun socket-connect (host port &key (protocol :stream) (element-type 'character)
+(defun socket-connect-internal (host &key port (protocol :stream) (element-type 'character)
                        timeout deadline
                        (nodelay t) ;; nodelay == t is the ACL default
                        local-host local-port)
@@ -129,8 +124,9 @@
       (t ; :io by default
        (iolib/sockets:shutdown (socket usocket) :read t :write t)))))
 
-(defun socket-listen (host port
-                           &key reuseaddress reuse-address
+(defun socket-listen-internal
+                          (host &key port reuseaddress
+                           reuse-address
                            (backlog 5)
                            (element-type 'character))
   (declare (ignore element-type))
@@ -227,8 +223,17 @@
       (iolib/streams:hangup ()
 	(funcall disconnector :close)))))
 
+#||
+Too few arguments in call to #<Compiled-function (:internal usocket::make-usocket-error-handler) (Non-Global)  #x3020018C678F>:
+2 arguments provided, at least 3 required.
+
+Backtrace:
+  0: ((:internal usocket::make-usocket-error-handler) 17511955605035 2)
+  1: (iolib/multiplex::%handle-one-fd #<event base, 1 FDs monitored, using: #<epoll(4) multiplexer> #x3020018C963D> (8 (:error :read :write)) 54337.518855526D0 nil 0.0D0)
+||#
+
 (defun make-usocket-error-handler (usocket disconnector)
-  (lambda (fd event exception)
+  (lambda (fd event &optional exception)
     (declare (ignore fd event exception))
     (handler-case
 	(setf (state usocket) nil)
@@ -288,3 +293,10 @@
     ;; close the event-base after use
     (unless (eq event-base *event-base*)
       (close event-base))))
+
+(defmethod socket-option ((socket usocket) option)
+  (iolib/sockets:socket-option (socket-stream socket) option))
+
+(defmethod (setf socket-option) (new-value (socket usocket) option)
+  (let ((form `(setf (iolib/sockets:socket-option ,(socket-stream socket) ,option) ,new-value)))
+    (eval (macroexpand-1 form))))
