@@ -57,26 +57,25 @@
           +unix-errno-error-map+))
 
 (defun raise-usock-err (errno socket &optional condition (host-or-ip nil))
+  (declare (ignore condition))
   (let ((usock-error
          (cdr (assoc errno +lispworks-error-map+ :test #'member))))
-    (if usock-error
-        (if (subtypep usock-error 'error)
-            (cond ((subtypep usock-error 'ns-error)
-                   (error usock-error :socket socket :host-or-ip host-or-ip))
-                  (t
-                   (error usock-error :socket socket)))
-            (cond ((subtypep usock-error 'ns-condition)
-                   (signal usock-error :socket socket :host-or-ip host-or-ip))
-                  (t
-                   (signal usock-error :socket socket))))
-      (error 'unknown-error
-             :socket socket
-             :real-error condition
-             :errno errno))))
+    (when usock-error
+      (if (subtypep usock-error 'error)
+          (cond ((subtypep usock-error 'ns-error)
+                 (error usock-error :socket socket :host-or-ip host-or-ip))
+                (t
+                 (error usock-error :socket socket)))
+        (cond ((subtypep usock-error 'ns-condition)
+               (signal usock-error :socket socket :host-or-ip host-or-ip))
+              (t
+               (signal usock-error :socket socket)))))))
 
 (defun handle-condition (condition &optional (socket nil) (host-or-ip nil))
   "Dispatch correct usocket condition."
   (typecase condition
+    (comm:socket-error (error 'socket-error :socket socket))
+    (conditions:stream-closed-error (error 'invalid-socket-stream-error :socket socket))
     (condition (let ((errno #-win32 (lw:errno-value)
                             #+win32 (wsa-get-last-error)))
                  (unless (zerop errno)
@@ -101,6 +100,16 @@
   (tv-sec :long)
   (tv-usec :long))
 
+(defconstant *sockopt_so_broadcast*
+  #-linux #x0020
+  #+linux 6
+  "Socket broadcast")
+
+(defconstant *sockopt_so_keepalive*
+  #-linux #x0008
+  #+linux 9
+  "Socket keepalive")
+
 ;;; ssize_t
 ;;; recvfrom(int socket, void *restrict buffer, size_t length, int flags,
 ;;;          struct sockaddr *restrict address, socklen_t *restrict address_len);
@@ -285,6 +294,60 @@
                                  len))
         zero-or-one 0))) ; on error, return 0
 
+(defun get-socket-broadcast (socket-fd)
+  "Get socket option: SO_BROADCAST, return value is a fixnum (0 or 1)"
+  (declare (type integer socket-fd))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int)
+                                     (len :int))
+    (if (zerop (comm::getsockopt socket-fd
+                                 comm::*sockopt_sol_socket*
+                                 *sockopt_so_broadcast*
+                                 (fli:copy-pointer zero-or-one
+                                                   :type '(:pointer #+win32 :char #-win32 :void))
+                                 len))
+        zero-or-one 0)))
+
+(defun set-socket-broadcast (socket-fd new-value)
+  "Set socket option: SO_BROADCAST, argument is a fixnum (0 or 1)"
+  (declare (type integer socket-fd)
+           (type (integer 0 1) new-value))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int))
+    (setf (fli:dereference zero-or-one) new-value)
+    (when (zerop (comm::setsockopt socket-fd
+                                   comm::*sockopt_sol_socket*
+                                   *sockopt_so_broadcast*
+                                   (fli:copy-pointer zero-or-one
+                                                     :type '(:pointer #+win32 :char #-win32 :void))
+                                   (fli:size-of :int)))
+        new-value)))
+
+(defun get-socket-keepalive (socket-fd)
+  "Get socket option: SO_KEEPALIVE, return value is a fixnum (0 or 1)"
+  (declare (type integer socket-fd))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int)
+                                     (len :int))
+    (if (zerop (comm::getsockopt socket-fd
+                                 comm::*sockopt_sol_socket*
+				 *sockopt_so_keepalive*
+                                 (fli:copy-pointer zero-or-one
+                                                   :type '(:pointer #+win32 :char #-win32 :void))
+                                 len))
+        zero-or-one 0)))
+
+(defun set-socket-keepalive (socket-fd new-value)
+  "Set socket option: SO_KEEPALIVE, argument is a fixnum (0 or 1)"
+  (declare (type integer socket-fd)
+           (type (integer 0 1) new-value))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int))
+    (setf (fli:dereference zero-or-one) new-value)
+    (when (zerop (comm::setsockopt socket-fd
+                                   comm::*sockopt_sol_socket*
+                                   *sockopt_so_keepalive*
+                                   (fli:copy-pointer zero-or-one
+                                                     :type '(:pointer #+win32 :char #-win32 :void))
+                                   (fli:size-of :int)))
+        new-value)))
+
 (defun initialize-dynamic-sockaddr (hostname service protocol &aux (original-hostname hostname))
   (declare (ignorable original-hostname))
   #+(or lispworks4 lispworks5 lispworks6.0)
@@ -321,8 +384,8 @@
                     server-addr
                     (if (eql family comm::*socket_af_inet*)
                         (fli:size-of '(:struct comm::sockaddr_in))
-			(fli:size-of '(:struct comm::sockaddr_in6))))))
-	:bad-host)))
+                        (fli:size-of '(:struct comm::sockaddr_in6))))))
+        :bad-host)))
 
 (defun open-udp-socket (&key local-address local-port read-timeout
                              (address-family comm::*socket_af_inet*))
@@ -363,7 +426,7 @@
                       (progn
                         (comm::close-socket socket-fd)
                         (error "cannot bind"))))))
-	    socket-fd))
+            socket-fd))
       (error "cannot create socket"))))
 
 (defun connect-to-udp-server (hostname service
@@ -379,6 +442,13 @@
                                         :local-port local-port
                                         :read-timeout read-timeout
                                         :address-family address-family)))
+	;; For broadcast addresses, the broadcast option has to be set prior to COMM:CONNECT
+	(when (eql address-family comm::*socket_af_inet*)
+	  (let ((last-octet-pos (position #\. hostname :from-end t)))
+	    ;; Kludgy check for most broadcast IPv4 addresses in absence of netmask.
+	    ;; This will still fail with networks smaller than /24
+	    (when (and last-octet-pos (string= "255" (subseq hostname (1+ last-octet-pos))))
+              (set-socket-broadcast socket-fd 1))))
         (if socket-fd
             (if (comm::connect socket-fd server-addr server-addr-length)
                 ;; success, return socket fd
@@ -389,7 +459,7 @@
                 (error "cannot connect")))
           (error "cannot create socket"))))))
 
-(defun socket-connect (host port &key (protocol :stream) (element-type 'base-char)
+(defun socket-connect-internal (host &key port (protocol :stream) (element-type 'base-char)
                        timeout deadline (nodelay t)
                        local-host local-port)
   ;; What's the meaning of this keyword?
@@ -453,8 +523,8 @@
                      :connected-p (and host port t))))
        usocket))))
 
-(defun socket-listen (host port
-                           &key reuseaddress
+(defun socket-listen-internal
+                          (host &key port reuseaddress
                            (reuse-address nil reuse-address-supplied-p)
                            (backlog 5)
                            (element-type 'base-char))
@@ -477,8 +547,8 @@
                    (error 'address-in-use-error)))))
     (make-stream-server-socket sock :element-type element-type)))
 
-;; Note: COMM::GET-FD-FROM-SOCKET contains addition socket wait operations, which
-;; should NOT be applied on socket FDs who have already been called on W-F-I,
+;; Note: COMM::GET-FD-FROM-SOCKET contains additional socket-wait operations, which
+;; should NOT be applied to socket FDs which has already been called on W-F-I,
 ;; so we have to check the %READY-P slot to decide if this waiting is necessary,
 ;; or SOCKET-ACCEPT will just hang. -- Chun Tian (binghe), May 1, 2011
 
@@ -490,26 +560,31 @@
                      (comm::get-fd-from-socket (socket usocket)))
                    #-win32
                    (comm::get-fd-from-socket (socket usocket))))
-         (stream (make-instance 'comm:socket-stream
-                                :socket socket
-                                :direction :io
-                                :element-type (or element-type
-                                                  (element-type usocket)))))
+         (stream (when socket
+                  (make-instance 'comm:socket-stream
+                                  :socket socket
+                                  :direction :io
+                                  :element-type (or element-type
+                                                    (element-type usocket))))))
     #+win32
     (when socket
       (setf (%ready-p usocket) nil))
-    (make-stream-socket :socket socket :stream stream)))
+
+    ;; It is possible that socket is NIL. Previously the call of make-stream-socket
+    ;; raised a usocket:invalid-socket-error
+    (when socket
+      (make-stream-socket :socket socket :stream stream))))
 
 ;; Sockets and their streams are different objects
 ;; close the stream in order to make sure buffers
 ;; are correctly flushed and the socket closed.
 (defmethod socket-close ((usocket stream-usocket))
   "Close socket."
-  (close (socket-stream usocket)))
+  (close (socket-stream usocket) :abort t)) ; see "sbcl.lisp" for (:abort t)
 
 (defmethod socket-close ((usocket usocket))
   (with-mapped-conditions (usocket)
-     (comm::close-socket (socket usocket))))
+    (comm::close-socket (socket usocket))))
 
 (defmethod socket-close :after ((socket datagram-usocket))
   "Additional socket-close method for datagram-usocket"
@@ -532,9 +607,9 @@
   (unless (member direction '(:input :output :io))
     (error 'invalid-argument-error))
   (let ((what (case direction
-		(:input +shutdown-read+)
-		(:output +shutdown-write+)
-		(:io +shutdown-read-write+))))
+                (:input +shutdown-read+)
+                (:output +shutdown-write+)
+                (:io +shutdown-read-write+))))
     (with-mapped-conditions (usocket)
       #-(or lispworks4 lispworks5 lispworks6) ; lispworks 7.0+
       (comm::shutdown (socket usocket) what)
@@ -545,13 +620,13 @@
   (unless (member direction '(:input :output :io))
     (error 'invalid-argument-error))
   (with-mapped-conditions (usocket)
-    #-(or lispworks4 lispworks5 lispworks6)
-    (comm:socket-stream-shutdown (socket usocket) direction)
+    #-(or lispworks4 lispworks5 lispworks6) ; lispworks 7+
+    (comm:socket-stream-shutdown (socket-stream usocket) direction)
     #+(or lispworks4 lispworks5 lispworks6)
     (let ((what (case direction
-		  (:input +shutdown-read+)
-		  (:output +shutdown-write+)
-		  (:io +shutdown-read-write+))))
+                  (:input +shutdown-read+)
+                  (:output +shutdown-write+)
+                  (:io +shutdown-read-write+))))
       (= 0 (%shutdown (comm:socket-stream-socket (socket usocket)) what)))))
 
 (defmethod initialize-instance :after ((socket datagram-usocket) &key)
@@ -593,9 +668,9 @@
           n
         (let ((errno #-win32 (lw:errno-value)
                      #+win32 (wsa-get-last-error)))
-          (if (zerop errno)
-              n
-            (raise-usock-err errno socket-fd host)))))))
+          (unless (zerop errno)
+            (raise-usock-err errno socket-fd host))
+          0)))))
 
 (defmethod socket-receive ((socket datagram-usocket) buffer length &key timeout (max-buffer-size +max-datagram-packet-size+))
   "Receive message from socket, read-timeout is a float number in seconds.
@@ -609,7 +684,7 @@
                    (integer 0)                          ; size
                    (unsigned-byte 32)                   ; host
                    (unsigned-byte 16))                  ; port
-	   (type sequence buffer))
+           (type sequence buffer))
   (let ((socket-fd (socket socket))
         (message (slot-value socket 'recv-buffer)) ; TODO: how multiple threads do this in parallel?
         (read-timeout timeout)
@@ -657,9 +732,9 @@
                                                            :copy-foreign-object nil)))
             (let ((errno #-win32 (lw:errno-value)
                          #+win32 (wsa-get-last-error)))
-              (if (zerop errno)
-                  (values nil n 0 0)
-                (raise-usock-err errno socket-fd)))))))))
+              (unless (zerop errno)
+                (raise-usock-err errno socket-fd))
+              (values nil n 0 0))))))))
 
 (defmethod get-local-name ((usocket usocket))
   (multiple-value-bind
@@ -691,9 +766,9 @@
     (setq hostname (comm:string-ip-address hostname))
     (unless hostname
       (let ((resolved-hostname (comm:get-host-entry hostname :fields '(:address))))
-	(unless resolved-hostname
-	  (return-from ipv6-address-p nil))
-	(setq hostname resolved-hostname))))
+        (unless resolved-hostname
+          (return-from ipv6-address-p nil))
+        (setq hostname resolved-hostname))))
   (comm:ipv6-address-p hostname))
 
 (defun lw-hbo-to-vector-quad (hbo)
@@ -749,19 +824,19 @@
       (dolist (x (wait-list-waiters wait-list))
         (mp:notice-fd (os-socket-handle x)))
       (labels ((wait-function (socks)
-		 (let (rv)
-		   (dolist (x socks rv)
-		     (when (usocket-listen x)
-		       (setf (state x) :READ
-			     rv t))))))
-	(if timeout
-	    (mp:process-wait-with-timeout "Waiting for a socket to become active"
-					(truncate timeout)
-					#'wait-function
-					(wait-list-waiters wait-list))
-	    (mp:process-wait "Waiting for a socket to become active"
-			     #'wait-function
-			     (wait-list-waiters wait-list))))
+                 (let (rv)
+                   (dolist (x socks rv)
+                     (when (usocket-listen x)
+                       (setf (state x) :READ
+                             rv t))))))
+        (if timeout
+            (mp:process-wait-with-timeout "Waiting for a socket to become active"
+                                        (truncate timeout)
+                                        #'wait-function
+                                        (wait-list-waiters wait-list))
+            (mp:process-wait "Waiting for a socket to become active"
+                             #'wait-function
+                             (wait-list-waiters wait-list))))
       (dolist (x (wait-list-waiters wait-list))
         (mp:unnotice-fd (os-socket-handle x)))
       wait-list))
