@@ -24,14 +24,14 @@
 
 (defparameter +openmcl-nameserver-error-map+
   '((:no-recovery . ns-no-recovery-error)
-    (:try-again . ns-try-again-condition)
+    (:try-again . ns-try-again-error)
     (:host-not-found . ns-host-not-found-error)))
 
 ;; we need something which the openmcl implementors 'forgot' to do:
 ;; wait for more than one socket-or-fd
 
 (defun input-available-p (sockets &optional ticks-to-wait)
-  (ccl::rletZ ((tv :timeval))
+  (ccl::rletz ((tv :timeval))
     (ccl::ticks-to-timeval ticks-to-wait tv)
     ;;### The trickery below can be moved to the wait-list now...
     (ccl::%stack-block ((infds ccl::*fd-set-size*))
@@ -58,11 +58,16 @@
         (error usock-err :socket socket)
       (error 'unknown-error :socket socket :real-error real-condition))))
 
-(defun handle-condition (condition &optional socket)
+(defun handle-condition (condition &optional socket (host-or-ip nil))
   (typecase condition
     (openmcl-socket:socket-error
        (raise-error-from-id (openmcl-socket:socket-error-identifier condition)
                             socket condition))
+    #+ccl-1.12
+    (ccl::stream-is-closed-error
+     (error 'invalid-socket-stream-error :socket socket))
+    (ccl:output-timeout
+       (error 'timeout-error :socket socket))
     (ccl:input-timeout
        (error 'timeout-error :socket socket))
     (ccl:communication-deadline-expired
@@ -73,26 +78,26 @@
 					    +openmcl-nameserver-error-map+))))
 	 (if nameserver-error
              (if (typep nameserver-error 'serious-condition)
-                 (error nameserver-error :host-or-ip nil)
-                 (signal nameserver-error :host-or-ip nil))
+                 (error nameserver-error :host-or-ip host-or-ip)
+                 (signal nameserver-error :host-or-ip host-or-ip))
 	   (raise-error-from-id condition-id socket condition))))))
 
 (defun to-format (element-type protocol)
   (cond ((null element-type)
 	 (ecase protocol ; default value of different protocol
-	   (:stream :text)
+	   (:stream ;;:text
+	     :bivalent)
 	   (:datagram :binary)))
 	((subtypep element-type 'character)
 	 :text)
-	(t :binary)))
+	(t ;; :binary
+	 :bivalent)))
 
 #-ipv6
-(defun socket-connect (host port &key (protocol :stream) element-type
-		       timeout deadline nodelay
+(defun socket-connect-internal (host &key port (protocol :stream) element-type
+		       timeout deadline (nodelay t)
 		       local-host local-port)
-  (when (eq nodelay :if-supported)
-    (setf nodelay t))
-  (with-mapped-conditions ()
+  (with-mapped-conditions (nil host)
     (ecase protocol
       (:stream
        (let ((mcl-sock
@@ -124,14 +129,14 @@
 	 usocket)))))
 
 #-ipv6
-(defun socket-listen (host port
-                      &key reuseaddress
+(defun socket-listen-internal
+                     (host &key port reuseaddress
                            (reuse-address nil reuse-address-supplied-p)
                            (backlog 5)
                            (element-type 'character))
   (let* ((reuseaddress (if reuse-address-supplied-p reuse-address reuseaddress))
 	 (real-host (host-to-hostname host))
-         (sock (with-mapped-conditions ()
+         (sock (with-mapped-conditions (nil host)
                   (apply #'openmcl-socket:make-socket
                          (append (list :connect :passive
                                        :reuse-address reuseaddress
@@ -153,15 +158,26 @@
 ;; by the same object.
 (defmethod socket-close ((usocket usocket))
   (with-mapped-conditions (usocket)
-    (close (socket usocket))))
+    (close (socket usocket) :abort t))) ; see "sbcl.lisp" for (:abort t)
 
 (defmethod socket-shutdown ((usocket usocket) direction)
-  (with-mapped-conditions (usocket)
-    (openmcl-socket:shutdown (socket usocket) :direction direction)))
+  (let ((ccl-direction (case direction
+                         (:io :both)
+                         (otherwise direction))))
+    (with-mapped-conditions (usocket)
+      (if (and (eq ccl-direction :both)
+               ;; The :BOTH support was introduced in CCL 1.11.5,
+               ;; and CCL's *FEATURES* allow to distignguish only major version,
+               ;; so we use it starting with 1.12.
+               (not (member :ccl-1.12 *features*)))
+          (progn
+            (openmcl-socket:shutdown (socket usocket) :direction :input)
+            (openmcl-socket:shutdown (socket usocket) :direction :output))
+          (openmcl-socket:shutdown (socket usocket) :direction ccl-direction)))))
 
 #-ipv6
 (defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0))
-  (with-mapped-conditions (usocket)
+  (with-mapped-conditions (usocket host)
     (if (and host port)
 	(openmcl-socket:send-to (socket usocket) buffer size
 				:remote-host (host-to-hbo host)
@@ -216,11 +232,11 @@
           (get-peer-port usocket)))
 
 (defun get-host-by-address (address)
-  (with-mapped-conditions ()
+  (with-mapped-conditions (nil address)
      (openmcl-socket:ipaddr-to-hostname (host-to-hbo address))))
 
 (defun get-hosts-by-name (name)
-  (with-mapped-conditions ()
+  (with-mapped-conditions (nil name)
      (list (hbo-to-vector-quad (openmcl-socket:lookup-hostname
                                 (host-to-hostname name))))))
 
