@@ -1,6 +1,4 @@
-;;;; $Id$
-;;;; $URL$
-
+;;;; -*- Mode: LISP; Base: 10; Syntax: ANSI-Common-lisp; Package: USOCKET -*-
 ;;;; See LICENSE for licensing information.
 
 (in-package :usocket)
@@ -26,11 +24,11 @@ to continue.
             :documentation "Indicates the minimal version of the
 implementation required to support the requested feature."))
   (:report (lambda (c stream)
-	     (format stream "~A in ~A is unsupported."
-		     (feature c) (context c))
-	     (when (minimum c)
-	       (format stream " Minimum version (~A) is required."
-		       (minimum c)))))
+             (format stream "~A in ~A is unsupported."
+                     (feature c) (context c))
+             (when (minimum c)
+               (format stream " Minimum version (~A) is required."
+                       (minimum c)))))
   (:documentation "Signalled when the underlying implementation
 doesn't allow supporting the requested feature.
 
@@ -39,8 +37,8 @@ When you see this error, go bug your vendor/implementation developer!"))
 (define-condition unimplemented (insufficient-implementation)
   ()
   (:report (lambda (c stream)
-	     (format stream "~A in ~A is unimplemented."
-		     (feature c) (context c))))
+             (format stream "~A in ~A is unimplemented."
+                     (feature c) (context c))))
   (:documentation "Signalled if a certain feature might be implemented,
 based on the features of the underlying implementation, but hasn't
 been implemented yet."))
@@ -49,8 +47,8 @@ been implemented yet."))
 
 (define-condition socket-condition (condition)
   ((socket :initarg :socket
-           :accessor usocket-socket))
-  ;;###FIXME: no slots (yet); should at least be the affected usocket...
+           :accessor usocket-socket
+           :documentation "Socket that raised the condition"))
   (:documentation "Parent condition for all socket related conditions."))
 
 (define-condition socket-error (socket-condition error)
@@ -63,14 +61,16 @@ been implemented yet."))
   (:documentation "Parent condition for all name resolution conditions."))
 
 (define-condition ns-error (ns-condition error)
-  ()
+  ((socket :initarg :socket
+           :accessor usocket-socket))
   (:documentation "Parent error for all name resolution errors."))
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defun define-usocket-condition-class (class &rest parents)
     `(progn
        (define-condition ,class ,parents ())
-       (export ',class))))
+       (eval-when (:load-toplevel :execute)
+         (export ',class)))))
 
 (defmacro define-usocket-condition-classes (class-list parents)
   `(progn ,@(mapcar #'(lambda (x)
@@ -89,11 +89,11 @@ been implemented yet."))
   (:documentation "Condition raised when there's no other - more applicable -
 condition available."))
 
-
 ;; Mass define and export our errors
 (define-usocket-condition-classes
   (address-in-use-error
    address-not-available-error
+   already-shutdown-error
    bad-file-descriptor-error
    connection-refused-error
    connection-aborted-error
@@ -109,6 +109,7 @@ condition available."))
    network-reset-error
    host-down-error
    host-unreachable-error
+   out-of-memory-error
    shutdown-error
    timeout-error
    deadline-timeout-error
@@ -116,6 +117,25 @@ condition available."))
    invalid-socket-stream-error)
   (socket-error))
 
+;; This is obsolated. USOCKET no more raises it inside HANDLE-CONDITION (from backends),
+;; so that code patterns like below becomes possible: (previously on SBCL, my-error was
+;; captured by HANDLE-CONDITION and was packed into UNKNOWN-ERROR whose :real-error is
+;; that my-error instance. See also https://github.com/usocket/usocket/issues/97
+#|
+(ql:quickload "usocket")
+
+(define-condition my-error (error) ())
+
+(handler-case
+    (usocket:with-client-socket (socket stream "google.com" 443
+                                        :element-type '(unsigned-byte 8))
+      ;; some my code
+      ;; ...
+      (error 'my-error ))
+  (my-error (c)
+    ;; handle my error
+             ))
+|#
 (define-condition unknown-error (socket-error)
   ((real-error :initarg :real-error
                :accessor usocket-real-error
@@ -137,8 +157,8 @@ condition available."))
 error available."))
 
 (define-usocket-condition-classes
-  (ns-try-again-condition)
-  (ns-condition))
+  (ns-try-again-condition) ; obsoleted
+  (socket-condition))
 
 (define-condition ns-unknown-condition (ns-condition)
   ((real-condition :initarg :real-condition
@@ -153,7 +173,8 @@ condition available."))
   ;; with lisp, we just return NIL (indicating no data) instead of
   ;; raising an exception...
   (ns-host-not-found-error
-   ns-no-recovery-error)
+   ns-no-recovery-error
+   ns-try-again-error)
   (ns-error))
 
 (define-condition ns-unknown-error (ns-error)
@@ -171,13 +192,17 @@ condition available."))
   (:documentation "Error raised when there's no other - more applicable -
 error available."))
 
-(defmacro with-mapped-conditions ((&optional socket) &body body)
-  `(handler-bind ((condition #'(lambda (c) (handle-condition c ,socket))))
-    ,@body))
+(defmacro with-mapped-conditions ((&optional socket host-or-ip) &body body)
+  "Run `body', handling implementation-specific conditions by re-raising them as usocket conditions.
+
+When `socket' or `host-or-ip' are specified, their values will be passed as arguments to the corresponding usocket conditions."
+  `(handler-bind ((condition
+                   #'(lambda (c) (handle-condition c ,socket ,host-or-ip))))
+     ,@body))
 
 (defparameter +unix-errno-condition-map+
-  `(((11) . ns-try-again-condition) ;; EAGAIN
-    ((35) . ns-try-again-condition) ;; EDEADLCK
+  `(((11) . ns-try-again-error)     ;; EAGAIN
+    ((35) . ns-try-again-error)     ;; EDEADLCK
     ((4) . interrupted-condition))) ;; EINTR
 
 (defparameter +unix-errno-error-map+
@@ -198,7 +223,7 @@ error available."))
     ((55 105) . no-buffers-error)
     ((12) . out-of-memory-error)
     ((45 95) . operation-not-supported-error)
-    ((1) . operation-not-permitted-error)
+    ((1 13) . operation-not-permitted-error)
     ((43 92) . protocol-not-supported-error)
     ((44 93) . socket-type-not-supported-error)
     ((51 101) . network-unreachable-error)
@@ -217,19 +242,18 @@ error available."))
 
 (defparameter +unix-ns-error-map+
   `((1 . ns-host-not-found-error)
-    (2 . ns-try-again-condition)
+    (2 . ns-try-again-error)
     (3 . ns-no-recovery-error)))
 
 (defmacro unsupported (feature context &key minimum)
   `(cerror "Ignore it and continue" 'unsupported
-	   :feature ,feature
-	   :context ,context
-	   :minimum ,minimum))
+           :feature ,feature
+           :context ,context
+           :minimum ,minimum))
 
 (defmacro unimplemented (feature context)
   `(signal 'unimplemented :feature ,feature :context ,context))
 
-
 ;;; People may want to ignore all unsupported warnings, here it is.
 (defmacro ignore-unsupported-warnings (&body body)
   `(handler-bind ((unsupported
