@@ -1,12 +1,36 @@
+;;;; See LICENSE for licensing information.
+
 (in-package :usocket)
 
+(defvar *server*)
+
 (defun socket-server (host port function &optional arguments
                       &key in-new-thread (protocol :stream)
                            ;; for udp
                            (timeout 1) (max-buffer-size +max-datagram-packet-size+)
                            ;; for tcp
-                           element-type reuse-address multi-threading
+                           element-type (reuse-address t) multi-threading
                            name)
+  "Create a simple TCP or UDP socket server.
+
+`host' and `port' name the local interface and port the server will listen on.
+`function' is the handler function. It must take at least one argument: a stream
+for a TCP server, or a buffer for a UDP server. For the UDP server, the handler
+must also return a buffer.
+`arguments' is a list of additional arguments to pass to the handler function,
+after the first.
+If `in-new-thread' is true, the server will be spawned on a new thread and this
+function will return immediately. Otherwise, this function will run
+indefinitely. `name' specifies the name of the thread.
+`protocol' can be :stream for a TCP server (the default) or :datagram for a UDP server.
+For UDP servers:
+`timeout' specifies a read timeout, 1 second by default.
+`max-buffer-size' specifies the maximum size a UDP packet can take up, `+max-datagram-packet-size+' by default.
+For TCP servers:
+`element-type' specifies the stream's element type.
+If `reuse-address' is true, wildcard hosts and more specific hosts can share a port.
+If `multi-threading' is true, each connection will be launched in a new thread,
+allowing the server to handle multiple connections in parallel."
   (let* ((real-host (or host *wildcard-host*))
          (socket (ecase protocol
                    (:stream
@@ -29,13 +53,25 @@
                                   :timeout timeout
                                   :max-buffer-size max-buffer-size)))))
       (if in-new-thread
-	  (values (portable-threads:spawn-thread (or name "USOCKET Server") #'real-call) socket)
-	  (real-call)))))
+          (values (bt2:make-thread #'real-call :name (or name "USOCKET Server")) socket)
+          (progn
+            (setq *server* socket)
+            (real-call))))))
 
 (defvar *remote-host*)
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (setf (documentation '*remote-host* 'variable)
+        "The remote host of a TCP or UDP event. This variable is dynamically bound~
+ in the context of a `socket-server', specifically the handler function."))
+
 (defvar *remote-port*)
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (setf (documentation '*remote-port* 'variable)
+        "The remote port of a TCP or UDP event. This variable is dynamically bound~
+ in the context of a `socket-server', specifically the handler function."))
 
 (defun default-udp-handler (buffer) ; echo
+  "Example handler for a UDP socket-server. Returns datagrams to sender."
   (declare (type (simple-array (unsigned-byte 8) *) buffer))
   buffer)
 
@@ -52,7 +88,7 @@
               (multiple-value-bind (recv n *remote-host* *remote-port*)
                   (socket-receive socket buffer max-buffer-size)
                 (declare (ignore recv))
-                (if (plusp n)
+                (if (<= 0 n)
                     (progn
                       (let ((reply
                              (apply function (subseq buffer 0 n) arguments)))
@@ -70,15 +106,17 @@
       (values))))
 
 (defun default-tcp-handler (stream) ; null
+  "Example handler for a TCP socket-server. Sends client 'Hello, world!' then closes the stream."
   (declare (type stream stream))
-  (terpri stream))
+  (format stream "Hello world!~%"))
 
 (defun echo-tcp-handler (stream)
+  "Example handler for a TCP socket-server, being an echo server."
   (loop
      (when (listen stream)
        (let ((line (read-line stream nil)))
-	 (write-line line stream)
-	 (force-output stream)))))
+     (write-line line stream)
+     (force-output stream)))))
 
 (defun tcp-event-loop (socket function &optional arguments
                        &key element-type multi-threading)
@@ -90,16 +128,28 @@
                              (socket-close client-socket)
                              nil))))
     (unwind-protect
-        (loop do
-          (let* ((client-socket (apply #'socket-accept
-                                       `(,socket ,@(when element-type `(:element-type ,element-type)))))
-                 (client-stream (socket-stream client-socket)))
-            (if multi-threading
-                (apply #'portable-threads:spawn-thread "USOCKET Client" real-function client-socket arguments)
-              (prog1 (apply real-function client-socket arguments)
-                (close client-stream)
-                (socket-close client-socket)))
-            #+scl (when thread:*quitting-lisp* (return))
-            #+(and cmu mp) (mp:process-yield)))
+         (loop do
+           (block continue
+             (let* ((client-socket (apply #'socket-accept
+                                          socket
+                                          (when element-type
+                                            (list :element-type element-type))))
+                    (client-stream (socket-stream client-socket)))
+                    (if multi-threading
+                        (bt2:make-thread
+                         (lambda ()
+                           (handler-case (apply real-function client-socket arguments)
+                             #+sbcl
+                             (sb-bsd-sockets:invalid-argument-error ())))
+                         :name "USOCKET Client")
+                        (unwind-protect
+                             (handler-case (apply real-function client-socket arguments)
+                               #+sbcl
+                               (sb-bsd-sockets:invalid-argument-error ()
+                                 (return-from continue)))
+                          (close client-stream)
+                          (socket-close client-socket)))
+                    #+scl (when thread:*quitting-lisp* (return))
+                    #+(and cmu mp) (mp:process-yield))))
       (socket-close socket)
       (values))))
